@@ -1,42 +1,58 @@
 use strict;
 package NetServer::Portal::Pi;
 use NetServer::Portal qw(term);
+use Cwd;
+use Symbol;
+use Scalar::Util qw(reftype);
 use Data::Dumper;
 
-# Lots of code is borrowed from Devel::Symdump.  (Thanks!)
-# clone Vim commands
-
 NetServer::Portal->register(cmd => "pi",
 			    title => "Perl Introspector",
 			    package => __PACKAGE__);
 
+if (!defined &Data::Dumper::Maxdepth) {
+    warn "Data::Dumper::Maxdepth is not available
+NetServer::Portal::Pi will be too verbose";
+    eval q[sub Data::Dumper::Maxdepth{}];
+}
+sub new_dumper {
+    my $o = Data::Dumper->new(@_);
+    $o->Indent(1);
+    $o;
+}
+
 my $Help = "*** Perl Introspector ***
 
 
 Type ':help' to load this buffer!
 
 output buffer commands:
-  ,        previous screen full
-  .        next screen full
-  :10      jump to line 10 (like vi)
-  /REx     search for /REx/
+  ,         previous screen full
+  .         next screen full
+  :10       jump to line 10 (like vi)
+  /REx      search for /REx/
 
-  <        previous buffer
-  >        next buffer
+  <         previous buffer
+  >         next buffer
 
-  :history show history of commands
+  :w <file> write buffer to <file>
+  :history  show history of commands
 
 navigational commands:
 
-  ls [-1fpv]
+  ls [-1afpv]
     -1 single column format
+    -a show everything
     -f functions
     -p sub-packages
-    -v variables (scalars, arrays, hashes, IOs)
+    -v variables (scalars, arrays, hashes, and IOs)
 
   cd ..          move to preceeding \"directory\"
   cd <package>   use 'ls' to see options
-  cd \$code       make evaluate to a REF
+  cd \$code       must evaluate to a REF
+                 (for fun, try: cd [1,2,[],4,[],6])
+
+  :pwd           show the working \"directory\"
 ";
 
 sub new {
@@ -67,7 +83,7 @@ sub update {
     # turn off line numbers with
     # :set number
 
-    my $output_rows = $rows - 2;
+    my $output_rows = $rows - 4;
     if (@{$o->{O}}) {
 	my $cur = $o->{O}[$o->{buffer}];
 	my $O = $cur->{buf};
@@ -83,6 +99,11 @@ sub update {
 	    $to > $#$O;
 	for (my $lx= $cur->{line}; $lx <= $to; $lx++) {
 	    my $l = $O->[$lx];
+
+	    # snarfed from Carp:
+	    $l =~ s/([\200-\377])/sprintf("M-%c",ord($1)&0177)/eg;
+	    $l =~ s/([\0-\37\177])/sprintf("^%c",ord($1)^64)/eg;
+
 	    my $note = '';
 	    if ($lx == $cur->{line} or $lx == $to) {
 		$note = " [".($lx+1)."]";
@@ -105,7 +126,12 @@ sub update {
     }
     $s .= "\n";
     $s .= $ln->($o->{error});
-    $s .= (join('::', @{$o->{Package}}) || 'main').' ';
+    my $at='';
+    if (@{$o->{Path}}) {
+	my @p = @{$o->{Path}};
+	$at = "\$at=$p[$#p] "
+    }
+    $s .= (join('::', @{$o->{Package}}) || 'main')." $at";
     $s;
 }
 
@@ -116,7 +142,6 @@ sub add_buffer {
 }
 
 sub cmd {
-    no strict 'refs';
     my ($o, $c, $in) = @_;
     
     my $conf = $c->conf;
@@ -169,9 +194,30 @@ sub cmd {
 	    $o->{O}[$o->{buffer}]{line} = $1 - $Rows/2; #XXX
 	} elsif ($in eq 'help') {
 	    $o->add_buffer([split /\n/, $Help]);
+	} elsif ($in =~ /^w \s+ (.+)$/x) {
+	    my $to = $1;
+	    $to = getcwd.'/'.$to if
+		$to !~ m,^/,;
+	    my $buf = $o->{O}[$o->{buffer}]{buf};
+	    my $fh = gensym;
+	    open $fh, ">$to" or return $o->{error} = "open $to: $!";
+	    my $char=0;
+	    for (@$buf) {
+		$char += 1+length $_;
+		print $fh $_."\n";
+	    }
+	    close $fh;
+	    $o->{error} = "$to ".(0+@$buf)." lines, $char characters";
 	} elsif ($in eq 'history') {
 	    my @tmp = @{$o->{I}};
 	    $o->add_buffer(\@tmp);
+	} elsif ($in eq 'pwd') {
+	    my $Dumper = new_dumper($o->{Path});
+	    $Dumper->Maxdepth(1);
+	    $o->add_buffer(["Current path:",
+			    '',
+			    @{$o->{Package}},
+			    split(/\n/, $Dumper->Dump)]);
 	} else {
 	    $o->{error} = "$in?  Try :help for a list of commands.";
 	}
@@ -181,7 +227,8 @@ sub cmd {
     push @{$o->{I}}, $in;
     shift @{$o->{I}} if @{$o->{I}} > 16;
 
-    if ($in =~ m/^ls (\s+ -[1apfv]+)? $/x) {
+    if (!@{$o->{Path}} and $in =~ m/^ls (\s+ -[1apfv]+)? $/x) {
+	# call resolve here XXX
 	my %f;
 	if ($1) {
 	    my $flags = $1;
@@ -192,51 +239,45 @@ sub cmd {
 	    ++$f{$_} for qw(p f v);
 	}
 	++$f{p} if keys %f == 0;
+
+	my $package = (join('::', @{$o->{Package}}) || 'main') . '::';
 	my @got;
-	my $pack = join('::', @{$o->{Package}}) || 'main';
 
-	# TODO:
-	#
-	# this code gives too many false positives
-	#
-	# cope with control chars
-
-	while (my ($key,$val) = each(%{*{"$pack\::"}})) {
-	    next if !defined $val;
-	    local(*ENTRY) = $val;
-            #### PACKAGE ####
-            if ($f{p} and defined *ENTRY{HASH} &&
-		$key =~ /::$/ && $key ne "main::")
-            {
-		push @got, $key;
+	no strict;
+	local *stab = *{$package};
+	while (my ($key,$val) = each(%stab)) {
+	    local(*entry) = $val;
+	    if ($f{v} and $key !~ /^_</ and defined $entry) {
+		push @got, "\$$key";
 	    }
-            #### FUNCTION ####
-            if ($f{f} and defined *ENTRY{CODE}) {
-		push @got, "\&$key";
-            }
-            #### SCALAR ####
-            if ($f{v} and defined *ENTRY{SCALAR}) {
-		push @got, "\$" . $key;
-            }
-            #### ARRAY ####
-            if ($f{v} and defined *ENTRY{ARRAY}) {
+	    if ($f{v} and $key !~ /^_</ and @entry) {
 		push @got, "\@$key";
-            }
-            #### HASH ####
-            if ($f{v} and defined *ENTRY{HASH} &&
-		$key !~ /::/)
-	    {
-		push @got, "\%$key";
-            }
-            #### IO #### (only 5.003_10+)
-	    if ($f{v} and defined *ENTRY{IO}){
-		push @got, $key;
+	    }
+	    if ($key ne "main::" && $key ne "DB::" && %entry
+		&& $key !~ /^_</
+		&& !($package eq __PACKAGE__ and $key eq "stab")) {
+
+		if ($key =~ /::$/) {
+		    push @got, $key
+			if $f{p};
+		} elsif ($f{v}) {
+		    push @got, "\%$key";
+		}
+	    }
+	    my $fileno;
+	    if ($f{v} and defined($fileno = fileno(*entry))) {
+		push @got, "$key=$fileno";
+	    }
+	    if ($f{f} and defined &entry) {
+		push @got, "\&$key";
 	    }
 	}
 	@got = sort { $a cmp $b } @got;
 	if (!$f{1}) {
 	    my $maxlen=0;
 	    for (@got) {
+		s/([\200-\377])/sprintf("M-%c",ord($1)&0177)/eg;
+		s/([\0-\37\177])/sprintf("^%c",ord($1)^64)/eg;
 		$maxlen = length $_
 		    if $maxlen < length $_;
 	    }
@@ -259,14 +300,26 @@ sub cmd {
 	}
 	$o->add_buffer(\@got);
 
+    } elsif (@{$o->{Path}} and $in eq 'ls') {
+	my @p = @{$o->{Path}};
+	
+	my $Dumper = new_dumper([ $p[$#p] ]);
+	$Dumper->Maxdepth(1);
+	$o->add_buffer([split /\n/,  $Dumper->Dump]);
+
+    } elsif ($in =~ m,^cd (\s+/)? $,x) {
+	@{$o->{Package}} = ();
+	@{$o->{Path}} = ();
+	
     } elsif ($in =~ s/^cd\s+//) {
-	if ($in eq '..') {
-	    pop @{$o->{Package}};
-	} elsif ($in =~ m/^ (\w+) (::)? $/x) {
-	    push @{$o->{Package}}, $1;
-	} else {
-	    $o->{error} = "cd $in: not implemented";
+	my @at = $o->resolve_path($in);
+	return if (@at == 1 and !defined $at[0]);
+	@{$o->{Package}} = ();
+	while (@at and !ref $at[0]) {
+	    push @{$o->{Package}}, shift(@at);
 	}
+	@{$o->{Path}} = @at;
+
     } else {
 	$in .= "\n" if $in !~ /\n$/;
 	my @warn;
@@ -276,12 +329,12 @@ sub cmd {
 	my $pack = join('::', @{$o->{Package}}) || 'main';
 	my @eval = eval "no strict;\n#line 1 \"input\"\npackage $pack;\n$in";
 	if ($@) {
-	    $o->add_buffer([split /\n/, "package $pack;\n$in\n---\n$@"]);
+	    $o->add_buffer([split /\n/, "package $pack;\n$in---\n$@"]);
 	} else {
 	    my $warns='';
 	    $warns = join('', @warn)."---\n"
 		if @warn;
-	    my $Dumper = Data::Dumper->new(\@eval);
+	    my $Dumper = new_dumper(\@eval);
 	    $o->add_buffer([split /\n/, "package $pack;\n$in---\n".$warns.
 			    $Dumper->Dump]);
 	}
@@ -289,4 +342,61 @@ sub cmd {
     pop @{$o->{O}} if @{$o->{O}} > 16;
 }
 
+sub resolve_path {
+    my ($o, $path) = @_;
+    my @at = (@{$o->{Package}}, @{$o->{Path}});
+    @at=()
+	if $path =~ s,^/,,;
+    if ($path =~ m,^[\w\.\:/-]+$,) {
+	my @step = split m'/+', $path;
+	for my $step (@step) {
+	    next if $step eq '.';
+	    if ($step eq '..') {
+		pop @at;
+	    } else {
+		if (@at and ref $at[$#at]) {
+		    my $at = $at[$#at];
+		    my $to;
+		    if (reftype $at eq 'ARRAY') {
+			$to = $at->[$step];
+		    } elsif (reftype $at eq 'HASH') {
+			$to = $at->[$step];
+		    } else {
+			$o->{error} = "Can't cd $step through '$at'.";
+			return @at;
+		    }
+		    if (!ref $to) {
+			$o->{error} = "Can't cd $step into '$to'.";
+			return @at;
+		    }
+		    push @at, $to;
+		} else {
+		    $step =~ s/::$//;
+		    push @at, $step;
+		}
+	    }
+	}
+	@at;
+    } else {
+	my $package = join('::', @{$o->{Package}}) || 'main';
+	my @r = eval "no strict;\n#line 1 \"input\"\npackage $package;\n$path";
+	if ($@) {
+	    $o->{error} = $@;
+	    return undef;
+	}
+	if (!@r) {
+	    $o->{error} = "Nothing there.";
+	} elsif (@r == 1) {
+	    if (ref $r[0]) {
+		return (@at, $r[0])
+	    } else {
+		$o->{error} = "Can't cd into '$r[0]'";
+		return undef;
+	    }
+	} else {
+	    return (@at, \@r);
+	}
+    }
+}
+
 1;
